% Copyright 2013, Yi Yu <yuy@comp.nus.edu.sg>
% This program is for Personal Music Video Generation based on Modeling Scene Moods and is free: 
% You can redistribute it and/or modify
% it under the terms of the GNU General Public License.


function videoInfo = getVisualFeat(mpgFiles, visualFeatPath)
% This function to extract the Keyframe and ColorHistogram
% mpgFiles  = dir([direc '\*.mpg']);

bCheckBoundary          = 1;
chGridSize              = 4;
simThres                = 0.9;
maxDurationInOneShot    = 30;


for mf=1:numel(mpgFiles)
    fprintf('Extracting visual features for video %d/%d : %s, ', mf, numel(mpgFiles), mpgFiles(mf).name);
    [oPath,fileName,fileExt] = fileparts(mpgFiles(mf).name);
    visualFeatMatFile = fullfile(visualFeatPath, [fileName,'.mat']);
    if exist(visualFeatMatFile, 'file')
        load(visualFeatMatFile);
        videoInfo(mf) = thisVideo;
        fprintf('skipped\n');
        continue;
    else
        fprintf('\n');
    end;
    
%     [pathstr, fname, ext, versn] = fileparts(mpgFiles(mf).name);
%     videodirec = [direc '\' fname];
%     if ~exist(videodirec, 'dir')
%         mymkdir (videodirec);
%     else
%         continue;
%     end;
%     
%    videoFileName = [direc '\' char(mpgFiles(mf).name)];
    
    % create the video object
    % ---------------------------------------------------------------------
    videoObj = mmreader(mpgFiles(mf).name);

    % get number of frames
    % ---------------------------------------------------------------------
    numFrames       = get(videoObj, 'numberOfFrames');
    frameRate       = get(videoObj, 'FrameRate');
    rows            = get(videoObj, 'Height');
    cols            = get(videoObj, 'Width');
    duration        = get(videoObj, 'Duration');
    maxFrameInShot  = ceil(maxDurationInOneShot*frameRate);
    numBins         = chGridSize*chGridSize*chGridSize;
    frameGap        = max(ceil(frameRate/2), 5); % capture at 2 frames per second    

    feaSpace        = zeros(maxFrameInShot, numBins);
    fridx           = zeros(maxFrameInShot, 1);

    % extract the boundary box
    % ---------------------------------------------------------------------  
    if bCheckBoundary == 1
        [row_start, row_end, col_start, col_end] = GetBoxBoundary(videoObj, numFrames, rows, cols);
    else
        row_start = 1;
        col_start = 1;
        row_end = rows;
        col_end = cols;
    end;
       
    % extract all frames
    % ---------------------------------------------------------------------
    numkeyframes = 1;
    idx = 1;
    frameidx = 1:frameGap:numFrames-1;
    if (frameidx(end) ~= numFrames-1)
        frameidx = [frameidx numFrames-1];
    end;
    
%     fid = fopen ([videodirec '\' fname '.bdy'], 'w');    
%     shot_start = 1;  

    visualFeat = zeros(floor(numFrames/15), numBins);
    keyframes  = zeros(floor(numFrames/15), 1);
    for f=1:numel(frameidx)-1
        fprintf('Extracting frames for video %d/%d : %s, frame_group %d/%d', mf, numel(mpgFiles), mpgFiles(mf).name, f, numel(frameidx)-1);

        % get the most stable frame for current time sample range
        % -----------------------------------------------------------------
        scope_idx = frameidx(f):2:frameidx(f+1)-1;
%         shot_end  = frameidx(f+1)-1;  
        scope_fea = zeros(length(scope_idx), numBins);
        for sc=1:length(scope_idx)
            scope_fea(sc,:) = getHist(read(videoObj, scope_idx(sc)), chGridSize);    
        end;
        stableFrameID = GetStableFrame (scope_fea);
        fr = scope_idx(stableFrameID);

        curFrame = read(videoObj, fr);
        % invalid frame
        if isempty(curFrame)
            fprintf('empty frame\n');
            continue;
        end;
        curFrame = curFrame(row_start:row_end, col_start:col_end,:);
        % get the color histogram for selected frame
        % -----------------------------------------------------------------
        feaSpace(idx, :)    = getHist(curFrame, chGridSize);
        fridx(idx)          = fr;

        % compute similarity value with the first image of current shot
        % -----------------------------------------------------------------
        simVal = sum(min(reshape(feaSpace(1,:),numBins,1), reshape(feaSpace(idx,:),numBins,1)));

        % shot boundary detected
        % -----------------------------------------------------------------
        if  (simVal < simThres   || idx == maxFrameInShot - 1) || fr == numFrames
            % find the best keyframe among the set of keyframes
            % -------------------------------------------------------------
            stableFrameID = GetStableFrame (feaSpace(1:max(idx-1, 1),:));            
            keyframes(numkeyframes) = fridx(stableFrameID);

            % output the stable frame
            % -------------------------------------------------------------
%             fn = sprintf('%010d', keyframes(numkeyframes));
%             fprintf(fid, '%s %f %d %d %d\n', fn, frameRate, shot_start, keyframes(numkeyframes), shot_end);
            curFrame = read(videoObj, keyframes(numkeyframes));

            %curFrame = RemoveBox (curFrame);  %HK, remove movie black box
            flg  = whetherdel(curFrame);      %RC, delete the frame contains too few values
            if ~flg
                [csind,ceind,rsind,reind,hqflg]  = removeBlkbox(curFrame);  %richard
                if hqflg
                    curFrame                     = curFrame(rsind:reind,csind:ceind,:);
%                     imgFileName                  = sprintf('%s\\%010d.jpg', videodirec, keyframes(numkeyframes));
%                     imwrite(curFrame, imgFileName);
                end
            end
            visualFeat(numkeyframes,:) = getHist(curFrame, chGridSize);

            % prepare for the next shot
            % -------------------------------------------------------------
            feaSpace(1,:) = feaSpace(idx,:);
            fridx(1)      = fridx(idx);
            idx = 2;
            fprintf (' -- frame %d \n', keyframes(numkeyframes));
            numkeyframes = numkeyframes + 1;            
%             shot_start = frameidx(f+1) + 1;
        else
            idx = idx + 1;
            fprintf('\n');
        end;     
    end;

    videoInfo(mf).visualFeat = visualFeat(1:numkeyframes-1,:);
    videoInfo(mf).keyframes = keyframes(1:numkeyframes-1);
    videoInfo(mf).frameTime = keyframes(1:numkeyframes-1)/frameRate;

    thisVideo = videoInfo(mf);
    save(visualFeatMatFile, 'thisVideo');
%     fclose(fid);
end;
% fclose(ifid);  %image files list
% 
% t = toc;
% fid = fopen([direc 'result.log'], 'w');
% fprintf(fid, 'Frame Converstion Time = %f\n', t);
% fclose(fid);
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% refine the frame by Richard
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function del = whetherdel(curframe)
        % delete according to the ratio of black pixels to white pixels
        del       = false;
        ratio_th  = 0.6; %the ratio, we care more informative keyframe, we can set this even 0.5.
        lvalue_th = 15;  %the threshold
        hvalue_th = 240;
        
        if length(size(curframe)) == 3
            curframe = mean(curframe,3);
        end
        
        lratio = length(find(curframe < lvalue_th)) / prod(size(curframe));
        hratio = length(find(curframe > hvalue_th)) / prod(size(curframe));
        if lratio > ratio_th || hratio > ratio_th 
            del = true;
        end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Exclude the black box by simple  operation by Richard
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [csind,ceind,rsind,reind,hqflg] = removeBlkbox(img)
    diff_th   = 10;
    hqflg     = true;   %high quality flag

    if length(size(img)) == 3
          img = mean(img,3);
    end
    
    % cut black column
    cdif = max(img,[],1) - min(img,[],1);
    csind = min(find(cdif > diff_th)); %column start index
    ceind = max(find(cdif > diff_th)); %column end index
    
    % cut black row
    rdif = max(img,[],2) - min(img,[],2);
    rsind = min(find(rdif > diff_th)); %row start index
    reind = max(find(rdif > diff_th)); %row end index
    
    if isempty(reind) || isempty(ceind) || isempty(csind) || isempty(rsind)
       hqflg = false;
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get the box
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [row_start, row_end, col_start, col_end] = GetBoxBoundary(videoObj, numFrames, rows, cols)

k = 30;
thres = 70;
frList1 = randperm(numFrames);
frList1 = frList1(1:min(numFrames, k));
frList2 = randperm(numFrames);
frList2 = frList2(1:min(numFrames, k));
avgDiff = zeros(rows, cols);

for i=1:k
    rawframe1 = read(videoObj, frList1(i)); 
    if size(rawframe1,2) > 2
        rawframe1 = double(rgb2gray(rawframe1));
    end;
    
    rawframe2 = read(videoObj, frList2(i));    
    if size(rawframe2,2) > 2
        rawframe2 = double(rgb2gray(rawframe2));
    end;
    
    diff = abs(rawframe1-rawframe2);
    if max(max(diff)) < 20
        continue;
    end;
    avgDiff = max(avgDiff,diff);
end

% vertical boundary
positions = find(max(avgDiff, [], 2) < thres);
if ~isempty(positions)
    if ~isempty(find(positions<rows/2))
        row_start = max(positions(positions<rows/2));
    else
        row_start = 1;
    end;
    if ~isempty(find(positions>rows/2))
        row_end   = min(positions(positions>rows/2));
    else
        row_end = rows;
    end;
end;

if isempty(positions) || row_end - row_start < rows*0.5
    row_start = 1;
    row_end = rows;
end;

% horizontal boundary
positions = find(max(avgDiff) < thres);
if ~isempty(positions)
    if ~isempty(find(positions<cols/2))
        col_start = max(positions(positions<cols/2));
    else
        col_start = 1;
    end;
    if ~isempty(find(positions>cols/2))
        col_end   = max(positions(positions>cols/2));
    else
        col_end = cols;
    end;
end;
if isempty(positions) || col_end - col_start < cols*0.6
    col_start = 1;
    col_end  = cols;        
end;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get the valid movie region
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function new_frame = RemoveBox(frame)

new_frame = [];

% get horizonal boundary
grFrame = double(rgb2gray(frame));
diff1 = mean(abs(grFrame - repmat(mean(grFrame,2), 1, size(grFrame,2))), 2);
st_idx = find (diff1>10);
if isempty(st_idx)
    return;
end;
row_start = st_idx(1);
row_end   = st_idx(length(st_idx));
if (row_end - row_start)/size(frame,1) < 0.5;
    row_start = 1;
    row_end = size(frame,1);
end;

% get vertical boundary
diff2 = mean(abs(grFrame - repmat(mean(grFrame), size(grFrame,1),1)));
st_idx = find(diff2>10);
if isempty(st_idx)
    return;
end;
col_start = st_idx(1);
col_end   = st_idx(length(st_idx));
if (col_end - col_start)/size(frame,2) < 0.5;
    col_start = 1;
    col_end = size(frame,2);
end;

% mainly blank screen
if (col_end - col_start)/size(frame, 2) < 0.25 || (row_end - row_start)/size(frame, 1) < 0.25
    row_start = -1;
    row_end   = -1;
    col_start = -1;
    col_end   = -1;    
    return;
end;

% valid output is expected
new_frame       = frame;
ann_row_end     = row_end;
chk_row_start = floor(row_start + (row_end - row_start)*0.6);
chk_row_end   = row_end;
diff1 = abs(double(frame(chk_row_start:chk_row_end-3, col_start:col_end,1)) - double(frame(chk_row_start+3:chk_row_end, col_start:col_end,1)));
diff2 = abs(double(frame(chk_row_start:chk_row_end-3, col_start:col_end,2)) - double(frame(chk_row_start+3:chk_row_end, col_start:col_end,2)));
diff3 = abs(double(frame(chk_row_start:chk_row_end-3, col_start:col_end,3)) - double(frame(chk_row_start+3:chk_row_end, col_start:col_end,3)));
diff = max(diff1, diff2);
diff = max(diff, diff3);
hor_line_array = find(sum(diff > 50,2)/(col_end - col_start) > 0.70);
if ~isempty(hor_line_array)
    ann_row_end = hor_line_array(1) + chk_row_start;
    if (ann_row_end - row_start)/size(frame,1) > 0.5
        row_end = ann_row_end;
    end;
end;

new_frame = frame(row_start:row_end, col_start:col_end, :);

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get the most stable frame among the set of features
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function stableFrameID = GetStableFrame (features)

frsim = zeros(size(features,1),1);
for i=1:size(features,1)
    for j=1:size(features,1)
        if i==j
            continue;
        end;
        frsim(i) = frsim(i) + sum(min(reshape(features(i,:),size(features,2),1), reshape(features(j,:),size(features,2),1)));
    end
end

[v i] = sort(frsim, 'descend');     
stableFrameID = i(1);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get Color Histogram
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function h = getHist(clip, bins)

clip = double(clip);

if (nargin <= 1)
   bins = 8;
end

z = size(clip,3);
clip2 = zeros(size(clip,1),size(clip,2));

f = 1;
for i = 1:z
   clip2 = clip2 + f*floor(clip(:,:,i)*bins/256);
   f=f*bins;
end

h = hist(reshape(clip2, size(clip2,1)*size(clip2,2), 1), 0:(f-1));
h = h / sum(h);
end