% Copyright 2013, Yi Yu <yuy@comp.nus.edu.sg>
% This program is for Personal Music Video Generation based on Modeling Scene Moods and is free: 
% You can redistribute it and/or modify
% it under the terms of the GNU General Public License.


clear all; close all;

config = getConfig;

load(config.geoTagMoodInfoFile);
% 'geoTagSet', 'moodTagSet'


% ----------- 4. build music database -------------------------------------
if ~exist(config.audioFeatFile, 'file')
    musicDbInfo = textread(config.musicDbFile, '%s', 'delimiter', '\n');
    numGenre = 0;
    musicGenreSet = [];
    for sIdx=1:numel(musicDbInfo)
        musicTrackInfo = regexp(musicDbInfo{sIdx}, ',', 'split');
        numGenre = numGenre + 1;
        musicGenreSet{numGenre} = lower(strtrim(musicTrackInfo{1}));
    end;
    musicGenreSet = unique(musicGenreSet);

    % per-audio-track processing
    for sIdx=1:numel(musicDbInfo)
        if mod(sIdx,20)==0
            fprintf('Audio feature extraction progress: %3d/%3d\n', sIdx, numel(musicDbInfo));
        end;
        if ismember(sIdx, [248])
            musicDb(sIdx).trackName = [];
            musicDb(sIdx).feat = [];
            musicDb(sIdx).genreIndex = [];
            musicDb(sIdx).moodScore = [];
            continue;
        end;
        musicTrackInfo = regexp(musicDbInfo{sIdx}, ',', 'split');
        fileName = strtrim(musicTrackInfo{6});
        fileName = [config.musicRoot  fileName(3:end)];
        [p,f,e]  = fileparts(fileName);
        tempFile = fullfile(config.tempDir, [f e]);
        thisAudioFeatFile = fullfile(config.tempDir, sprintf('audio_%04d.mat', sIdx));
        if ~exist(thisAudioFeatFile, 'file')
            copyfile(fileName, tempFile, 'f');
            thisAudioFeat = getAudioFeat(tempFile);
            delete(tempFile);
            save(thisAudioFeatFile, 'thisAudioFeat');
        else
            load(thisAudioFeatFile);
        end;
        musicDb(sIdx).trackName = fileName;
        musicDb(sIdx).feat = thisAudioFeat;

        genreInfo = lower(strtrim(musicTrackInfo{1}));
        index = strcmp(musicGenreSet, genreInfo);
        musicDb(sIdx).genreIndex = find(index);

        moodTagStr  = musicTrackInfo{19};
        moodTagInfo = regexp(moodTagStr, ';', 'split');
        moodScore = zeros(numel(moodTagInfo),2);
        for k=1:numel(moodTagInfo)
            index = strcmp(moodTagSet, lower(strtrim(moodTagInfo{k})));
            moodIndex = find(index);
            if isempty(moodIndex)
                moodIndex = 0;
            end;
            moodScore(k,1) = moodIndex;
            moodScore(k,2) = 1/k;
        end;
        musicDb(sIdx).moodScore = moodScore;
    end;
    save(config.audioFeatFile, 'musicGenreSet', 'musicDb');
else
    load(config.audioFeatFile);
end;

if ~exist(config.audioBucketFile, 'file')
    for mIdx=1:numel(moodTagSet)
        musicBucket(mIdx).bucket= [];
    end;

    musicGenreInfo = [];
    for gIdx=1:numel(musicGenreSet)
        musicIdxPerGenre = find([musicDb(:).genreIndex]==gIdx);
        % use 60% of songs as the music database
        numTrack = floor(numel(musicIdxPerGenre)*0.6);
        for sIdx=1:numTrack
            musicIdx = musicIdxPerGenre(sIdx);
            for mIdx=1:size(musicDb(musicIdx).moodScore,1)
                moodIndex = musicDb(musicIdx).moodScore(mIdx,1);
                if moodIndex==0
                    continue;
                end;
                if moodIndex>12
                    fprintf('debug musicIdx=%d\n', musicIdx);
                end;
                moodScore = musicDb(musicIdx).moodScore(mIdx,2);
                musicBucket(moodIndex).bucket = [musicBucket(moodIndex).bucket; 
                                                [musicIdx moodScore]];
            end;
        end;
        musicGenreInfo(gIdx).allTrack = musicIdxPerGenre;
        musicGenreInfo(gIdx).trackInDb= musicIdxPerGenre(1:numTrack);
        musicGenreInfo(gIdx).trackInQr= musicIdxPerGenre(numTrack+1:end);
    end;
    save(config.audioBucketFile, 'musicBucket', 'musicGenreInfo');
else
    load(config.audioBucketFile);
end;
