% Copyright 2013, Yi Yu <yuy@comp.nus.edu.sg>
% This program is for Personal Music Video Generation based on Modeling Scene Moods and is free: 
% You can redistribute it and/or modify
% it under the terms of the GNU General Public License.


clear all; close all;

config = getConfig;

load(config.geoTagMoodInfoFile);
% 'geoTagSet', 'moodTagSet'

load(config.geoFeatFile);
% 'annotationInfo().geoTagFeat'
% 'annotationInfo().startTime'
% 'annotationInfo().endTime'

load(config.visualFeatFile);
% 'visualFeat().visualFeat'
% 'visualFeat().frameTime'

load(config.audioFeatFile);
load(config.audioBucketFile);
% 'musicGenreSet', 'musicDb', 'musicBucket', 'musicGenreInfo'


% ----------- 5. music recommendation -------------------------------------
if 1
    % 5.1 get features of video, user listening history
%     % test 1
%     videoIndex = 321;  % annotationInfo().geoTagFeat, visualFeat().visualFeat
%     genreIndex = 1;
%     listenHistory = musicGenreInfo(genreIndex).trackInQr([1 2 3]); % musicDb().feat

%     % test 2
%     videoIndex = 331;  % annotationInfo().geoTagFeat, visualFeat().visualFeat
%     genreIndex = 2;
%     listenHistory = musicGenreInfo(genreIndex).trackInQr([1 2 3]); % musicDb().feat

    % test 3
    videoIndex = 352;  % annotationInfo().geoTagFeat, visualFeat().visualFeat
    genreIndex = 3;
    listenHistory = musicGenreInfo(genreIndex).trackInQr([1 2 3]); % musicDb().feat

    
    fprintf('Recommend music for [%d]th video out of [%d] videos\n', videoIndex, numel(visualFeat));
    fprintf('    User listening history genre is [%s], music index = [', musicGenreSet{genreIndex});
    fprintf(repmat('%d,',1,numel(listenHistory)), listenHistory);
    fprintf('] out of [%d] songs\n', numel(musicDb));
    
    % synchronization between geo-feature and visual feature
    fIdx = videoIndex;
    numKeyFrm = find(visualFeat(fIdx).frameTime <= annotationInfo(fIdx).endTime(end), 1, 'last');
    geoTagFeat = zeros(numKeyFrm, size(annotationInfo(fIdx).geoTagFeat,2));
    moodIndex  = zeros(numKeyFrm, 1);
    numItem = 0;
    for sIdx=1:numel(annotationInfo(fIdx).startTime)
        index = visualFeat(fIdx).frameTime >= annotationInfo(fIdx).startTime(sIdx) ...
              & visualFeat(fIdx).frameTime <  annotationInfo(fIdx).endTime(sIdx);
        numThisSeg = sum(index);
        geoTagFeat(numItem+1:numItem+numThisSeg,:) = repmat(annotationInfo(fIdx).geoTagFeat(sIdx,:), numThisSeg, 1);
        moodIndex( numItem+1:numItem+numThisSeg,:) = repmat(annotationInfo(fIdx).moodIndex( sIdx,:), numThisSeg, 1);
        numItem = numItem + numThisSeg;
    end;
    thisVisualFeat = visualFeat(fIdx).visualFeat(1:numKeyFrm,:);

    for aIdx=1:numel(listenHistory)
        audioIdx = listenHistory(aIdx);
        audioFeat(aIdx) = musicDb(audioIdx);
    end;
else
    % 5.1 extract feature
    videoFile;  % => visualFeat
    geoFile;    % => geoFeat
    listenFile; % => audioFeat
end;


% 5.2 get the trained model
[emitWeight, transWeight] = getSvmModelInfo(config.trainModelFile);
svmhmm.emitW = emitWeight;
svmhmm.transW= transWeight;


% 5.3 find 'uniqMood' (top-3?) with 'geoTagFeat', 'thisVisualFeat';
extFeat = [];
if config.useGeoFeat
    extFeat = [extFeat geoTagFeat];
end;
if config.useVisualFeat
    extFeat = [extFeat thisVisualFeat];
end;
recogMood = viterbi_svm_Nbest(svmhmm, extFeat);
histInfo = getHistInfo(recogMood);
numFoundMood = min(size(histInfo,1), 3);
uniqMood = histInfo(1:numFoundMood,1);
fprintf('Recognized moods for the video: ');
for k=1:numFoundMood
    fprintf('%s', moodTagSet{uniqMood(k)});
end;
fprintf('\n');

% 5.4 get 'songInfo' with 'uniqMood'
candiSongs = [];
for mIdx=1:numel(uniqMood)
    moodIndex = uniqMood(mIdx);
    candiSongs = [candiSongs; musicBucket(moodIndex).bucket];
end;
uniqSong = unique(candiSongs(:,1));
songInfo = zeros(numel(uniqSong),2);
songInfo(:,1) = uniqSong;
for k=1:numel(uniqSong)
    index = candiSongs(:,1)==uniqSong(k);
    songInfo(k,2) = sum(candiSongs(index,2));
end;

% 5.5 compare feat of 'songInfo' with 'audioFeat' (user listening history)
similarity = zeros(size(songInfo,1),numel(audioFeat));
for aIdx=1:numel(audioFeat)
    for sIdx=1:size(songInfo,1)
        audioIdx = songInfo(sIdx,1);
        similarity(sIdx,aIdx) = getAudioSimilarity(audioFeat(aIdx).feat, musicDb(audioIdx).feat);
    end;
end;
% finalScore = [songInfo(:,1), sum(similarity,2) .* songInfo(:,2)];
finalScore = [songInfo(:,1), sum(similarity,2)];
finalScore = sortrows(finalScore,2);
finalScore = finalScore(end:-1:1,:);

% final result: songIdxSet
numRecommendSong = min(size(finalScore,1),3);
fprintf('Recommended songs:\n');
recommend.songIdxSet = finalScore(1:numRecommendSong,1);
for k=1:numRecommendSong
    audioIdx = recommend.songIdxSet(k);
    recommend.trackName{k} = musicDb(audioIdx).trackName;
    annoMoodTagStr = '';
    for n=1:size(musicDb(audioIdx).moodScore,1)
        if musicDb(audioIdx).moodScore(n,1)~=0
            annoMoodTagStr = [annoMoodTagStr  moodTagSet{musicDb(audioIdx).moodScore(n,1)} ','];
        end;
    end;
    fprintf('    SongIdx=%3d, genre=%s, annotatedMood=%s\n        SongName=%s\n', ...
        audioIdx, musicGenreSet{musicDb(audioIdx).genreIndex}, ...
        annoMoodTagStr, recommend.trackName{k});
end;
