% Copyright 2013, Yi Yu <yuy@comp.nus.edu.sg>
% This program is for Personal Music Video Generation based on Modeling Scene Moods and is free: 
% You can redistribute it and/or modify
% it under the terms of the GNU General Public License.


clear all; close all;

config = getConfig;

% ----------- 1. get 'geoTagSet', 'moodTagSet' ----------------------------
if ~exist(config.geoTagMoodInfoFile, 'file')
    [geoTagSet, moodTagSet] = getAnnotationSet(config.annotationListFile, config.usePrimaryMoodOnly);
    save(config.geoTagMoodInfoFile, 'geoTagSet', 'moodTagSet');
else
    load(config.geoTagMoodInfoFile);
end;


% ----------- 2. feature extraction ---------------------------------------
% 2.1 extract geo-category feature
if ~exist(config.geoFeatFile, 'file')
    annotationInfo = getAnnotationInfo(config.annotationListFile, geoTagSet, moodTagSet, config.usePrimaryMoodOnly);
    save(config.geoFeatFile, 'annotationInfo');
else
    load(config.geoFeatFile);
end;
% annotationInfo(fIdx).geoTagFeat;
% annotationInfo(fIdx).moodIndex;
% annotationInfo(fIdx).startTime;
% annotationInfo(fIdx).endTime;
% annotationInfo(fIdx).videoFile;

% for fIdx = 1:numel(annotationInfo)
%     moodIndex = annotationInfo(fIdx).moodIndex;
%     fprintf(['V%03d:'  repmat('%3d,',1, numel(moodIndex)) '\n'], fIdx, moodIndex);
% end;


% 2.2 extract visual feature: color-histogram
if ~exist(config.visualFeatFile, 'file')
    for fIdx=1:numel(annotationInfo)
        videoFiles(fIdx).name = annotationInfo(fIdx).videoFile;
    end;
    visualFeat = getVisualFeat(videoFiles, tempDir);
    save(config.visualFeatFile, 'visualFeat');
else
    load(config.visualFeatFile);
end;
% visualFeat(fIdx).visualFeat = visualFeat(1:numkeyframes-1,:);
% visualFeat(fIdx).keyframes  = keyframes(1:numkeyframes-1);
% visualFeat(fIdx).frameTime  = keyframes(1:numkeyframes-1)/frameRate;


% 2.3 synchronization between geo-feature and visual feature
for fIdx=1:numel(annotationInfo)
    if config.useVisualFeat
        numKeyFrm = find(visualFeat(fIdx).frameTime <= annotationInfo(fIdx).endTime(end), 1, 'last');
        geoTagFeat = zeros(numKeyFrm, size(annotationInfo(fIdx).geoTagFeat,2));
        moodIndex  = zeros(numKeyFrm, 1);
        numItem = 0;
        for sIdx=1:numel(annotationInfo(fIdx).startTime)
            index = visualFeat(fIdx).frameTime >= annotationInfo(fIdx).startTime(sIdx) ...
                  & visualFeat(fIdx).frameTime <  annotationInfo(fIdx).endTime(sIdx);
            numThisSeg = sum(index);
            geoTagFeat(numItem+1:numItem+numThisSeg,:) = repmat(annotationInfo(fIdx).geoTagFeat(sIdx,:), numThisSeg, 1);
            moodIndex( numItem+1:numItem+numThisSeg,:) = repmat(annotationInfo(fIdx).moodIndex( sIdx,:), numThisSeg, 1);
            numItem = numItem + numThisSeg;
        end;
        geoFeat(fIdx).geoTagFeat = geoTagFeat;
        geoFeat(fIdx).moodIndex  = moodIndex;
        visualFeat(fIdx).visualFeat = visualFeat(fIdx).visualFeat(1:numKeyFrm,:);
        visualFeat(fIdx).keyframes  = visualFeat(fIdx).keyframes(1:numKeyFrm);
        visualFeat(fIdx).frameTime  = visualFeat(fIdx).frameTime(1:numKeyFrm);
        if size(geoTagFeat,1)~=numKeyFrm
            fprintf(2, 'fIdx=%3d, geoFeat=%2d, videoFeat=%2d\n', fIdx, size(geoTagFeat,1), numKeyFrm);
        end;
    elseif config.useGeoFeat % use geoFeat only
        geoFeat(fIdx).geoTagFeat = annotationInfo(fIdx).geoTagFeat;
        geoFeat(fIdx).moodIndex  = annotationInfo(fIdx).moodIndex;
    end;
end;
% return;


% ----------- 3. svm-hmm --------------------------------------------------
% 3.1 select training/test set
numClass = numel(moodTagSet);
% numTrain = floor(numel(visualFeat)*0.8);
% trainSet = 1:numTrain;
% testSet  = numTrain+1:numel(visualFeat);

moodForTrain = [];
for fIdx = 1:numel(annotationInfo) % numel(trainSet);
    moodForTrain = [moodForTrain; geoFeat(fIdx).moodIndex];
end;
moodHist = hist(moodForTrain, 1:numClass);
% plot(moodHist);

trainSet = [ % by manual selection
    01  05  07  11  16  17  18  19  20  22  23  31  35  37  ...
    38  40  45  46  47  48  51  52  53  54  73  79  85  98  ...
   101 103 110 111 112 116 121 129 137 138 149 152 153 155  ...
   162 163 167 197 280 281 282 292];
% plus automatic selection
allVideoSet = 1:numel(geoFeat);
otherVideoSet = setdiff(allVideoSet, trainSet);
numExtraTrain = floor(numel(geoFeat)*0.8 - numel(trainSet));
trainSet = [trainSet otherVideoSet(1:numExtraTrain)];
trainSet = sort(trainSet);
testSet = setdiff(allVideoSet, trainSet);

% moodForTrain = [];
% for trainIdx = 1:length(trainSet);
%     videoIdx = trainSet(trainIdx);
%     moodForTrain = [moodForTrain; unique(geoFeat(videoIdx).moodIndex)];
% end;
% moodHist = hist(moodForTrain, 1:numClass);
% plot(moodHist);


% 3.2 training a model if not existing
if ~exist(config.trainModelFile, 'file')
    for trainIdx = 1:length(trainSet);
        videoIdx = trainSet(trainIdx);
        extFeat = [];
        if config.useGeoFeat
            extFeat = [extFeat geoFeat(videoIdx).geoTagFeat];
        end;
        if config.useVisualFeat
            extFeat = [extFeat visualFeat(videoIdx).visualFeat];
        end;
        X_train{trainIdx} = extFeat;
        Y_train{trainIdx} = geoFeat(videoIdx).moodIndex;
    end;
    svmhmmwrite(config.trainDataFile,X_train,Y_train);
    clear X_train Y_train; %free up memory

    %string for learning model
    C = 400; epsilon = 0.0001;
    % C = 100; epsilon = 0.0001;
    trainCmd = ['svm_hmm\svm_hmm_learn.exe -c ' num2str(C) ' -e ' num2str(epsilon) ' ' config.trainDataFile ' ' config.trainModelFile];
    trainCmd  %so we can see if it looks ok
    %learn model
    dos(trainCmd);
end;

% 3.3 get the trained model
[emitWeight, transWeight] = getSvmModelInfo(config.trainModelFile);
svmhmm.emitW = emitWeight;
svmhmm.transW= transWeight;

% 3.4 test
fprintf('Testing the system\n');
numCorrect = 0;
for testIdx=1:numel(testSet)
    videoIdx = testSet(testIdx);
    extFeat = [];
    if config.useGeoFeat
        extFeat = [extFeat geoFeat(videoIdx).geoTagFeat];
    end;
    if config.useVisualFeat
        extFeat = [extFeat visualFeat(videoIdx).visualFeat];
    end;
    trueMood = geoFeat(videoIdx).moodIndex';
    trueMoodCnt = getHistInfo(trueMood);
    trueMoodStr = sprintf('%2d(%2d)', trueMoodCnt');

    recogMood = viterbi_svm_Nbest(svmhmm, extFeat);
    recogMoodCnt = getHistInfo(recogMood);
    typicalMood = recogMoodCnt(1,1);
    if typicalMood == trueMoodCnt(1,1)
        numCorrect= numCorrect+1;
        resultStr = 'OK';
    else
        resultStr = 'NG';
    end;
    fprintf('%3d, rate= %3d/%3d, %s, recogMood=%2d, trueMood=%s\n', ...
        videoIdx, sum(trueMood==recogMood), numel(trueMood), resultStr, typicalMood, trueMoodStr);
end;
fprintf('Total recog result = %3d / %3d\n', numCorrect, numel(testSet));


